/*
 * A test of executing code in new tasks and threads.
 *
 * Copyright (C) 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>

#include <fsclient/client.h>
#include <fsclient/process.h>
#include <systypes/fcntl.h>
#include <systypes/format.h>

#include <stdio.h>



/* Transfer size for communication. */

const offset_t TO_TRANSFER = 1024;



/* Test process initiation and interaction. */

static long test_process(int argc, const char *argv[])
{
  process_t *process;
  file_t *reader, *writer;
  long err;

  /* Obtain the common notifier. */

  notifier_t *notifier = client_notifier_task();

  /* Create a pipe for process output. */

  err = client_pipe(&reader, &writer, O_NONBLOCK);

  if (err)
  {
    printf("Could not obtain pipe: %s\n", l4sys_errtostr(err));
    return err;
  }

  err = client_subscribe(reader, NOTIFY_CONTENT_AVAILABLE | NOTIFY_PEER_CLOSED, notifier);

  if (err)
  {
    printf("Could not subscribe to pipe notifications: %s\n", l4sys_errtostr(err));
    return err;
  }

  /* Start the process. */

  err = process_spawn(argc, argv, writer, &process);

  if (err)
  {
    printf("Could not start process: %s\n", l4sys_errtostr(err));
    return err;
  }

  printf("Finished program initiation.\n");

  /* Release the writing end of the pipe. */

  client_close(writer);

  /* Wait for a signal from the process or input from the process. */

  err = notify_subscribe(process_notifiable(process), NOTIFY_TASK_ALL, notifier);

  if (err)
  {
    printf("Could not subscribe to task notifications: %s\n", l4sys_errtostr(err));
    return err;
  }

  notifiable_t *notifiable;

  while (1)
  {
    err = notify_wait_many(&notifiable, notifier);

    if (err)
    {
      printf("Could not wait for process: %s\n", l4sys_errtostr(err));
      return err;
    }

    /* Handle any signal. */

    if (notifiable == process_notifiable(process))
      break;

    /* Handle any input. */

    else if (notifiable == file_notifiable(reader))
    {
      char buffer[TO_TRANSFER];
      offset_t nread = client_read(reader, buffer, TO_TRANSFER);

      while (nread)
      {
        fwrite(buffer, sizeof(char), nread, stdout);
        nread = client_read(reader, buffer, TO_TRANSFER);
      }
    }
  }

  notify_flags_t flags = process_notifications(process);
  notify_values_t values = process_notification_values(process);

  printf("End process (flags %" pFMTnotify_flags "x values: %ld, %ld)\n", flags, values.sig, values.val);
  printf("End reader (flags %" pFMTnotify_flags "x)\n", file_notifications(reader));

  err = process_error(process);

  client_close(reader);
  process_free(process);
  client_notifier_close(notifier);

  return err;
}



int main(int argc, char *argv[])
{
  long err;

  if (argc < 2)
  {
    printf("Need the program to run.\n");
    return 1;
  }

  /* Start a process for a non-existent program. */

  printf("Start non-existent program...\n");

  const char *bad_argv[] = {"non_existent_program"};

  err = test_process(1, bad_argv);

  if (!err)
  {
    printf("Non-existent program was apparently successfully started.\n");
    return 1;
  }

  printf("Non-existent program result: %s\n", l4sys_errtostr(err));

  /* Start a process for the given program. */

  printf("Start process...\n");

  err = test_process(argc - 1, (const char **) argv + 1);

  if (err)
    return 1;

  printf("End of test.\n");
  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
